# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""TankPurchaseCostAlgorithm
"""
import biosteam as bst
import thermosteam as tmo
import flexsolve as flx
from biosteam.units.decorators import cost, copy_algorithm
from biosteam.units.design_tools import CEPCI_by_year, cylinder_diameter_from_volume, cylinder_area
from biosteam import tank_factory
import numpy as np

__all__ = (
    'GrainHandling',
    'HammerMill', 
    'CornStorage',
    'CleaningSystem',
    'MilledCornSurgeTank',
    'MilledCornHopper',
    'MilledCornWeighTank',
    'LimeHopper',
    'AmmoniaTank',
    'AlphaAmylaseTank',
    'SlurryMixTank',
    'Liquefaction',
    'JetCooker',
    'CookedSlurrySurgeTank',
    'GlucoAmylaseTank',
    'SulfuricAcidTank',
    'Saccharification',
    'YeastTank',
    'WetDDGSConveyor',
    'Liquefaction',
    'SimultaneousSaccharificationFermentation', 'SSF',
    'DDGSHandling',
    'DDGSCentrifuge',
    'PlantAir_CIP_WasteWater_Facilities',
)

CE2007 = CEPCI_by_year[2007]

CAS_water = '7732-18-5'
# Material factors. Applicable to some units only.
MF38 = 0.38 
MF90 = 0.90
MF52 = 0.52

@cost('Flow rate', units='kg/hr', CE=CE2007, cost=MF38 * 120800, S=46350.72, kW=97, n=0.6)
class GrainHandling(bst.Unit): pass

CornStorage = tank_factory('CornStorage', 
    CE=CE2007, cost=MF38 * 979300., S=185400, tau=259.2, n=1.0, V_wf=0.9, V_max=3e5, 
    V_units='m3'
)

@cost('Flow rate', units='kg/hr', CE=CE2007, cost=60300., S=45350., n=0.6, ub=7.2e5)
class CleaningSystem(bst.Splitter): pass
    

@cost('Flow rate', units='kg/hr', CE=CE2007, cost=MF38 * 98200., S=46211.33, n=0.6, ub=720000., kW=314.237)
class HammerMill(bst.Unit): pass


MilledCornSurgeTank = tank_factory('MilledCornSurgeTank', 
    CE=CE2007, cost=MF38 * 32500., S=76.90, tau=2.0, n=0.6, V_wf=0.9, V_max=200., V_units='m3'
)

MilledCornHopper = tank_factory('MilledCornHopper', 
    CE=CE2007, cost=50700., S=100.93, tau=2.0, n=0.6, V_wf=0.9, V_max=1e3, V_units='m3'
)

MilledCornWeighTank = tank_factory('MilledCornWeighTank',
    CE=CE2007, cost=MF38 * 43600., S=76.90, tau=2.0, n=0.6, V_wf=0.9, V_max=20e3, V_units='m3'
)

LimeHopper = tank_factory('LimeHopper', 
    CE=CE2007, cost=9100., S=4.02, tau=46.3, n=0.6, V_wf=0.75, V_max=100., V_units='m3'
)

AmmoniaTank = tank_factory('AmmoniaTank', 
    CE=CE2007, cost=MF38 * 28400., S=8.77, tau=100., n=0.6, V_wf=0.90, V_max=100., V_units='m3'
)

AlphaAmylaseTank = tank_factory('AlphaAmylaseTank', 
    CE=CE2007, cost=49700., S=12.77, tau=336., n=0.6, V_wf=0.90, V_max=100., V_units='m3'
)

SlurryMixTank = tank_factory('SlurryMixTank', mixing=True,
    CE=CE2007, cost=133300., S=45.29, tau=0.25, n=0.6, V_wf=0.65, V_max=80., kW_per_m3=1.1607597, V_units='m3',
)


@cost('Flow rate', units='kg/hr', CE=CE2007, cost=14000, n=0.6, S=150347)
class JetCooker(bst.Unit):
    """
    ins : stream sequence
    
        [0] Feed
        
        [1] Steam
    
    outs : stream
        Mixed product.
    
    """
    _N_outs = 1
    _N_ins = 2
    _N_heat_utilities = 0
    
    def __init__(self, ID="", ins=None, outs=(), thermo=None, T=483.15):
        super().__init__(ID, ins, outs, thermo)
        self.T = T
    
    @staticmethod
    def _T_objective_function(steam_mol, T, steam, effluent, feed):
        steam.imol[CAS_water] = abs(steam_mol)
        effluent.mol[:] = steam.mol + feed.mol
        effluent.H = feed.H + steam.H
        return effluent.T - T
    
    def _run(self):
        feed, steam = self._ins
        steam_mol = feed.F_mol / 100.
        effluent, = self.outs
        effluent.T = self.T
        steam_mol = flx.aitken_secant(self._T_objective_function,
                                      steam_mol, 1/8 * steam_mol + 1., 
                                      1e-4, 1e-4,
                                      args=(self.T, steam, effluent, feed),
                                      checkroot=False)
        effluent.P = steam.P / 2.

CookedSlurrySurgeTank = tank_factory('CookedSlurrySurgeTank',
    CE=CE2007, cost=MF90 * 173700., S=14.16, tau=0.25, n=0.6, V_wf=0.90, V_max=100., V_units='m3',
)

GlucoAmylaseTank = tank_factory('AlphaAmylaseTank', 
    CE=CE2007, cost=84200., S=17.57, tau=336., n=0.6, V_wf=0.90, V_max=100., V_units='m3'
)

SulfuricAcidTank = tank_factory('AlphaAmylaseTank', 
    CE=CE2007, cost=MF38 * 19300., S=18.87, tau=336., n=0.6, V_wf=0.90, V_max=283.17, V_units='m3'
)

Saccharification = tank_factory('AlphaAmylaseTank', kW_per_m3=0.036, mixing=True,
    CE=CE2007, cost=MF90 * 102700., S=52.10, tau=1./3., n=0.6, V_wf=0.90, V_min=20., V_max=610., V_units='m3'
)

YeastTank = tank_factory('YeastTank', kW_per_m3=0.5,
    CE=CE2007, cost=114700., S=2.97, tau=40., n=0.6, V_wf=0.90, V_max=80., V_units='m3'
)

@cost('Flow rate', S=24158., n=0.6, units='kg/hr', CE=CE2007, kW=13.1, cost=55700.)
class WetDDGSConveyor(bst.Unit): pass

LiquefactionTank = tank_factory('LiquefactionTank', 
    CE=CE2007, cost=160900., S=141.3, tau=0.9, n=0.6, V_wf=0.90, V_max=500., kW_per_m3=0.6,
    mixing=True,
)

class Liquefaction(LiquefactionTank):
    """
    Create a Liquefaction unit operation that models the converion
    for Starch to Glucose oligomers.
    
    Parameters
    ----------
    ins : stream
        Inlet fluid.
    outs : stream
        Outlet fluid.
    yield_: float
        Yield of starch to glucose as a fraction of the theoretical yield.
    
    Notes
    -----
    The conversion of Starch to Glucose oligomers is modeled according to the
    following stoichiometry:
        
    Starch + H2O -> Glucose
    
    Where starch is a chemical with formula C6H10O5 that represents linked 
    glucose monomers (dehydrated from linkage).
    
    The dextrose equivalent, and for that manner the degree of polymerization,
    is not taken into account in this unit. However, the conversion is equivalent
    to the conversion of starch to fermentable saccharides, which is what matters
    downstream.
    
    References
    ----------
    TODO
    
    """
    def __init__(self, *args, yield_=1.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.reaction = tmo.reaction.Reaction('Starch + H2O -> Glucose', 'Starch', yield_)
        
    @property
    def yield_(self):
        return self.reaction.X
    @yield_.setter
    def yield_(self, X):
        self.reaction.X = X

    def _run(self):
        effluent, = self.outs
        effluent.mix_from(self.ins)
        self.reaction(effluent)
        

class SimultaneousSaccharificationFermentation(bst.BatchBioreactor):
    """
    Create a SimultaneousSaccharificationFermentation unit operation that 
    models the simultaneous saccharification and fermentation in the conventional
    dry-grind enthanol process.
    
    Parameters
    ----------
    ins : streams
        Inlet fluids.
    outs : stream
        Outlet fluid.
    yield_: float
        Yield of glucose to ethanol as a fraction of the theoretical yield.
    
    Notes
    -----
    This unit operation doesn't actually model the saccharification process.
    The reactor is modeled by the stoichiometric conversion of glucose to
    ethanol by mol:
        
    .. math:: 
        Glucose -> 2Ethanol + 2CO_2
    
    Yeast is assumed to be produced from the any remaining glucose:
        Glucoes -> Yeast
    
    A compound with name 'Yeast' must be present. Note that only glucose is 
    taken into account for conversion. Cleaning and unloading time,
    `tau_0`, fraction of working volume, `V_wf`, and number of reactors,
    `N_reactors`, are attributes that can be changed. Cost of a reactor
    is based on the NREL batch fermentation tank cost assuming volumetric
    scaling with a 6/10th exponent [1]_. 
    
    References
    ----------
    .. [1] D. Humbird, R. Davis, L. Tao, C. Kinchin, D. Hsu, and A. Aden
        National. Renewable Energy Laboratory Golden, Colorado. P. Schoen,
        J. Lukas, B. Olthof, M. Worley, D. Sexton, and D. Dudgeon. Harris Group
        Inc. Seattle, Washington and Atlanta, Georgia. Process Design and Economics
        for Biochemical Conversion of Lignocellulosic Biomass to Ethanol Dilute-Acid
        Pretreatment and Enzymatic Hydrolysis of Corn Stover. May 2011. Technical
        Report NREL/TP-5100-47764
    
    
    """
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, 
                 tau=60.,  N=None, V=None, T=305.15, P=101325., Nmin=2, Nmax=36,
                 yield_=0.95, V_wf=0.83):
        bst.BatchBioreactor.__init__(self, ID, ins, outs, thermo,
            tau=tau, N=N, V=V, T=T, P=P, Nmin=Nmin, Nmax=Nmax
        )
        self.reaction = tmo.reaction.Reaction('Glucose -> 2Ethanol + 2CO2',  'Glucose', yield_)
        self.V_wf = V_wf
    
    def _run(self):
        vent, effluent = self.outs
        effluent.mix_from(self.ins)
        self.reaction(effluent)
        effluent.imass['Yeast'] += effluent.imass['Glucose', 'NH3'].sum()
        effluent.imass['Glucose', 'NH3'] = 0.
        vent.receive_vent(effluent)
    
SSF = SimultaneousSaccharificationFermentation

@copy_algorithm(bst.SolidLiquidsSplitCentrifuge, run=False)
class DDGSCentrifuge(bst.Splitter): pass
    

@cost('Flow rate', units='kg/hr', CE=CE2007, cost=122800, S=15303.5346, kW=37.3, n=0.6)
class DDGSHandling(bst.Unit): pass


class PlantAir_CIP_WasteWater_Facilities(bst.Facility):
    network_priority = 0
    
    def __init__(self, ID, corn):
        self.corn = corn
        super().__init__(ID)
        
    def _run(self):
        pass
        
    def _cost(self):
        C = self.baseline_purchase_costs
        C['Facilities'] = 6e5 * (self.corn.F_mass / 46211.6723)**0.6


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 11:05:24 2017

@author: Yoel
"""
import numpy as np
import biosteam as bst
import flexsolve as flx
from biosteam import units, SystemFactory
from biosteam import main_flowsheet as f

__all__ = (
    'create_feedstock_handling_system',
    'create_juicing_system_up_to_clarification',
    'create_juicing_system_with_fiber_screener',
    'create_sucrose_fermentation_system',
    'create_beer_distillation_system',
    'create_ethanol_purification_system_after_beer_column',
    'create_sucrose_to_ethanol_system',
    'create_ethanol_purification_system',
    'create_sugarcane_to_ethanol_system',
    'create_bagasse_pelleting_system',
)


# %% Juicing and evaporation

@SystemFactory(
    ID='bagasse_pelleting_sys',
    ins=[dict(ID='bagasse')],
    outs=[dict(ID='bagasse_pellets')]
)
def create_bagasse_pelleting_system(ins, outs):
    bagasse, = ins
    bagasse_pellets, = outs
    U401 = units.HammerMill('U401', bagasse)
    U402 = units.DrumDryer('U402', 
        (U401-0, 'dryer_air', 'dryer_natural_gas'), 
        ('', 'dryer_outlet_air', 'dryer_emissions'),
        moisture_content=0.15, split=0.,
    )
    # X401 = bst.ThermalOxidizer('X401', (U403-1, 'oxidizer_air'), 'oxidizer_emissions')
    U403 = units.ScrewFeeder('U403', U402-0)
    U404 = units.BagassePelletMill('U404', U403-0)
    U405 = units.ConveyingBelt('U405', U404-0, bagasse_pellets)

@SystemFactory(
    ID='feedstock_handling_sys',
    ins=[dict(ID='sugarcane',
              Water=0.7,
              Glucose=0.01208,
              Sucrose=0.1369,
              Ash=0.006,
              Cellulose=0.06115,
              Hemicellulose=0.03608,
              Lignin=0.03276,
              Solids=0.015,
              total_flow=333334.2,
              units='kg/hr',
              price=0.03455)],
    outs=[dict(ID='shredded_cane')]
)
def create_feedstock_handling_system(ins, outs):
    sugarcane, = ins
    shreaded_cane, = outs
    U101 = units.ConveyingBelt('U101', sugarcane)
    U102 = units.MagneticSeparator('U102', U101-0)
    U103 = units.Shredder('U103', U102-0, shreaded_cane)

@SystemFactory(
    ID='juicing_sys',
    ins=[create_feedstock_handling_system.ins[0],
         dict(ID='enzyme',
              Cellulose=100,
              Water=900,
              units='kg/hr',
              price=0.5),],
    outs=[dict(ID='untreated_juice'),
          dict(ID='bagasse')]
)
def create_juicing_system_without_treatment(ins, outs, pellet_bagasse=None):
    if pellet_bagasse is None: pellet_bagasse = True
    
    ### Streams ###
    sugarcane, enzyme = ins
    untreated_juice, bagasse = outs
    
    imbibition_water = bst.Stream('imbibition_water',
                                  Water=87023.35, units='kg/hr',
                                  T = 350.15)
    
    # Hydrolyze starch
    T201 = units.EnzymeTreatment('T201', (sugarcane, enzyme), T=323.15)  # T=50
    
    @T201.add_specification(run=True)
    def update_enzyme_and_imbibition_water():
        F_mass = T201.ins[0].F_mass
        enzyme.imass['Cellulose', 'Water'] = 0.003 * F_mass * np.array([0.1, 0.9])
        imbibition_water.imass['Water'] = 0.25 * F_mass
    
    # Finely crush lipid cane
    U201 = units.CrushingMill('U201',
                              ins=(T201-0, ''),
                              split=dict(Ash=0.92,
                                         Cellulose=0.92,
                                         Glucose=0.04,
                                         Hemicellulose=0.92,
                                         Lignin=0.92,
                                         Sucrose=0.04,
                                         Solids=1),
                              moisture_content=0.5)
    U202 = units.ConveyingBelt('U202', U201-0, [''] if pellet_bagasse else [bagasse])
    
    if pellet_bagasse:
        bagasse_pelleting_sys = create_bagasse_pelleting_system(None, ins=U202-0, outs=bagasse, mockup=True)
    
    # Mix in water
    M201 = units.Mixer('M201', ('', imbibition_water), 1-U201)
    
    # Screen out fibers
    S201 = units.VibratingScreen('S201', U201-1, ('', 0-M201),
                                 split=dict(Ash=0.35,
                                            Cellulose=0.35,
                                            Glucose=0.88,
                                            Hemicellulose=0.35,
                                            Lignin=0.35,
                                            Solids=0,
                                            Sucrose=0.88,
                                            Water=0.88))
    
    # Store juice before treatment
    T202 = units.StorageTank('T202', S201-0, untreated_juice,
                             tau=4, vessel_material='Carbon steel')


@SystemFactory(
    ID='juicing_sys',
    ins=[*create_juicing_system_without_treatment.ins,
         dict(ID='H3PO4',
              H3PO4=74.23,
              Water=13.1,
              units='kg/hr',
              price=0),
         dict(ID='lime',
              CaO=333.0,
              Water=2200.0,
              units='kg/hr',
              price=0.077),
         dict(ID='polymer',
              Flocculant=0.83,
              units='kg/hr',
              price=0)],
    outs=[dict(ID='clarified_juice'),
          dict(ID='bagasse')]
)
def create_juicing_system_up_to_clarification(ins, outs, pellet_bagasse=None):
    
    ### Streams ###
    sugarcane, enzyme, H3PO4, lime, polymer = ins
    clarified_juice, bagasse = outs
    
    rvf_wash_water = bst.Stream('rvf_wash_water',
                                Water=16770, units='kg/hr',
                                T=363.15)  # to C202
    
    juicing_sys_without_treatment = create_juicing_system_without_treatment(
        ins=(sugarcane, enzyme),
        outs=('', bagasse),
        mockup=True,
        pellet_bagasse=pellet_bagasse,
    )
    T201 = f.unit.T201
    
    # Heat up before adding acid
    H201 = units.HXutility('H201', juicing_sys_without_treatment-0, T=343.15)
    
    # Mix in acid
    T203 = units.MixTank('T203', (H201-0, H3PO4))
    
    # Pump acid solution
    P201 = units.Pump('P201', T203-0)
    
    # Mix lime solution
    T204 = units.MixTank('T204', lime, tau=0.10)
    
    # Blend acid lipid solution with lime
    T205 = units.MixTank('T205', (P201-0, T204-0), tau=0.10)
    P202 = units.Pump('P202', T205-0)
    
    # Mix recycle
    M202 = units.Mixer('M202', (P202-0, ''))
    
    # Heat before adding flocculant
    H202 = units.HXutility('H202', M202-0, T=372.15)
    
    # Mix in flocculant
    T206 = units.MixTank('T206', (H202-0, polymer))
    T206.tau = 0.10
    
    # Separate residual solids
    C201 = units.Clarifier('C201', T206-0, (clarified_juice, ''),
                           split=dict(Ash=0,
                                      CaO=0,
                                      Cellulose=0,
                                      Flocculant=0.522,
                                      Glucose=0.522,
                                      Hemicellulose=0,
                                      Lignin=0,
                                      H3PO4=0.522,
                                      Sucrose=0.522,
                                      Water=0.522))
    
    # Remove solids as filter cake
    C202 = units.RVF('C202', (C201-1, rvf_wash_water), ('filter_cake', ''),
                     moisture_content=0.80,
                     split=dict(Ash=0.85,
                                CaO=0.85,
                                Cellulose=0.85,
                                Glucose=0.01,
                                Hemicellulose=0.85,
                                Lignin=0.85,
                                Sucrose=0.01))
    P203 = units.Pump('P203', C202-1, 1-M202)
    
    ### Process specifications ###
    
    # Specifications dependent on lipid cane flow rate
    @T201.add_specification(run=True)
    def correct_flows():
        F_mass = T201.ins[0].F_mass
        # correct enzyme, lime, phosphoric acid, and imbibition water
        enzyme.F_mass *= F_mass / 333334.2
        lime.imass['CaO', 'Water'] = 0.001 * F_mass * np.array([0.046, 0.954])
        H3PO4.imass['H3PO4', 'Water'] = 0.00025 * F_mass
        T201._run()
    
    # Specifications within a system
    def correct_wash_water():
        P202._run()
        solids = P202.outs[0].imol['Ash', 'CaO', 'Cellulose',
                                   'Hemicellulose', 'Lignin'].sum()
        rvf_wash_water.imol['Water'] = 0.0574 * solids
    
    P202.specification = correct_wash_water

@SystemFactory(
    ID='juicing_sys',
    ins=create_juicing_system_up_to_clarification.ins,
    outs=[dict(ID='screened_juice'),
          dict(ID='bagasse'),
          dict(ID='fiber_fines')]
)          
def create_juicing_system_with_fiber_screener(ins, outs, pellet_bagasse=None):
    screened_juice, bagasse, fiber_fines = outs
    sys = create_juicing_system_up_to_clarification(
        None, ins, ['', bagasse],
        mockup=True,
        pellet_bagasse=pellet_bagasse
    )
    # Screen out small fibers from sugar stream
    S202 = units.VibratingScreen('S202', sys-0, (screened_juice, fiber_fines),
                                 split=dict(Ash=0.998,
                                            CaO=0.998,
                                            Cellulose=0.0,
                                            Flocculant=0.0,
                                            Glucose=0.998,
                                            Hemicellulose=0.0,
                                            Lignin=0.0,
                                            H3PO4=1.0,
                                            Sucrose=0.998,
                                            Water=0.998))
    S202.mesh_opening = 2


# %% Ethanol separation

@SystemFactory(
    ID='beer_distillation_sys',
    ins=[dict(ID='beer',
              T=348.34,
              P=101325,
              Water=96500.0,
              Ethanol=22550.0,
              Glucose=4916,
              H3PO4=83.33,
              Yeast=103,
              units='kg/hr')],
    outs=[dict(ID='distilled_beer'),
          dict(ID='stillage')] # TODO: Find good selling price for stillage/vinasse and possibly yeast
)
def create_beer_distillation_system(ins, outs,
                                    beer_column_heat_integration=True,
                                    IDs={}):
    beer, = ins
    distilled_beer, stillage = outs
    
    P301 = units.Pump(IDs.get('Beer pump', 'P301'), P=2. * 101325)
    
    # Beer column
    x_bot = 3.91e-06
    y_top = 0.28
    D302 = units.BinaryDistillation(IDs.get('Beer column', 'D302'), P=2. * 101325,
                                    outs=(distilled_beer, ''),
                                y_top=y_top, x_bot=x_bot, k=1.1, Rmin=0.001,
                                LHK=('Ethanol', 'Water'))
    D302.tray_material = 'Stainless steel 304'
    D302.vessel_material = 'Stainless steel 304'
    D302.boiler.U = 1.85
    P302 = units.Pump(IDs.get('Beer column bottoms product pump', 'P302'), P=101325.)
    
    # Heat up before beer column
    # Exchange heat with stillage    
    if beer_column_heat_integration:
        H302 = units.HXprocess(IDs.get('Beer column heat exchange', 'H302'), 
                               outs=('', stillage),
                               phase0='l', phase1='l', U=1.28)
        (beer-P301-0, P302-0)-H302-0-D302-1-P302
    else:
        beer-P301-0-D302-1-P302
        P302.outs[0] = stillage


@SystemFactory(
    ID='ethanol_purification_from_distilled_beer_sys',
    ins=[dict(ID='distilled_beer'),
         dict(ID='denaturant',
              Octane=230.69,
              units='kg/hr',
              price=0.756)],
    outs=[dict(ID='ethanol',
               price=0.789),
          dict(ID='stripper_bottoms_product')]
)
def create_ethanol_purification_system_after_beer_column(ins, outs, IDs={}):
    distilled_beer, denaturant = ins
    ethanol, stripper_bottoms_product = outs
  
    
    # Mix ethanol Recycle (Set-up)
    M303 = units.Mixer(IDs.get('Recycle mixer', 'M303'))
    
    D303 = units.BinaryDistillation(IDs.get('Distillation', 'D303'), 
                                x_bot=3.9106e-06, y_top=0.80805, k=1.25, Rmin=0.01,
                                LHK=('Ethanol', 'Water'),
                                tray_material='Stainless steel 304',
                                vessel_material='Stainless steel 304',
                                P=1013250.,
                                is_divided=True)
    D303.boiler.U = 1.85
    P303 = units.Pump(IDs.get('Distillation bottoms product pump', 'P303'), 
                      outs=stripper_bottoms_product)
    
    # Superheat vapor for mol sieve
    H303 = units.HXutility(IDs.get('Heat exchanger to superheat vapor to molecular sieves', 'H303'),
                           T=115+273.15, V=1, heat_only=True)
    
    # Molecular sieve
    U301 = units.MolecularSieve(IDs.get('Molecular sieves', 'U301'),
                                split=(2165.14/13356.04, 1280.06/1383.85),
                                order=('Ethanol', 'Water'))
    
    # Condense ethanol product
    H304 = units.HXutility(IDs.get('Ethanol condenser', 'H304'), 'S149', V=0, T=340.)
    T302 = units.StorageTank(IDs.get('Ethanol day tank', 'T302'), tau=12,
                             vessel_type='Floating roof',
                             vessel_material='Carbon steel')
    P304 = units.Pump(IDs.get('Ethanol day tank pump', 'P304'))
    
    # Storage for gasoline
    T303 = units.StorageTank(IDs.get('Denaturant storage', 'T303'), tau=7*24,
                             vessel_type='Floating roof',
                             vessel_material='Carbon steel')
    P305 = units.Pump(IDs.get('Denaturant pump', 'P305'))
    
    # Denatured ethanol product
    M304 = units.Mixer(IDs.get('Ethanol-denaturant mixer', 'M304'))
    T304 = units.StorageTank(IDs.get('Product tank', 'T304'),
                             vessel_type='Floating roof',
                             vessel_material='Carbon steel',
                             tau=6.5*24, outs=ethanol)
    
    ### Ethanol system set-up ###
    (distilled_beer, U301-0)-M303-0-D303-0-H303-U301
    D303-1-P303
    
    @P304.add_specification(run=True)
    def adjust_denaturant():
        pure_ethanol = P304.ins[0]
        denaturant.imol['Octane'] = 0.022*pure_ethanol.F_mass/114.232
    
    (denaturant-T303-P305-0, U301-1-H304-0-T302-0-P304-0)-M304-T304

# @SystemFactory(
#     ID='ethanol_purification_from_distilled_beer_sys',
#     ins=[dict(ID='distilled_beer'),
#           dict(ID='denaturant', Octane=230.69, units='kg/hr', price=0.756)],
#     outs=[dict(ID='ethanol', price=0.789),
#           dict(ID='stripper_bottoms_product')]
# )
# def create_ethanol_purification_system_after_beer_column(ins, outs):
#     distilled_beer, denaturant = ins
#     ethanol, stripper_bottoms_product = outs
#     M303 = units.Mixer('M303')
#     D303 = units.BinaryDistillation('D303',
#         P=101325, x_bot=3.9106e-06, y_top=0.80805, k=1.25, Rmin=0.01,
#         LHK=('Ethanol', 'Water'), tray_material='Stainless steel 304',
#         vessel_material='Stainless steel 304', is_divided=True
#     )
#     D303.boiler.U = 1.85
#     P303 = units.Pump('P303', outs=stripper_bottoms_product)
#     H303 = units.HXutility('H303', T=115+273.15, V=1)
#     U301 = units.MolecularSieve('U301', split=dict(Ethanol=0.162, Water=0.925))
#     H304 = units.HXutility('H304', V=0, T=340.)
#     T302 = units.StorageTank('T302', 
#         tau=12, vessel_type='Floating roof', vessel_material='Carbon steel'
#     )
#     P304 = units.Pump('P304')
#     T303 = units.StorageTank('T303', 
#         tau=7*24, vessel_type='Floating roof', vessel_material='Carbon steel'
#     )
#     P305 = units.Pump('P305'); M304 = units.Mixer('M304')
#     T304 = units.StorageTank('T304',
#         vessel_type='Floating roof', vessel_material='Carbon steel',
#         tau=6.5*24, outs=ethanol
#     )
#     def adjust_denaturant():
#         P304._run()
#         denaturant.imol['Octane'] = 0.022 * P304.outs[0].F_mass / 114.232
    
#     P304.specification = adjust_denaturant
#     (distilled_beer, U301-0)-M303-0-D303-0-H303-U301
#     D303-1-P303
#     (denaturant-T303-P305, U301-1-H304-0-T302-0-P304-0)-M304-T304


@SystemFactory(
    ID='ethanol_purification_sys',
    ins=[dict(ID='beer',
              T=348.34,
              P=101325,
              Water=96500.0,
              Ethanol=22550.0,
              Glucose=4916,
              H3PO4=83.33,
              Yeast=103,
              units='kg/hr'),
         create_ethanol_purification_system_after_beer_column.ins[1]],
    outs=[create_ethanol_purification_system_after_beer_column.outs[0],
          dict(ID='stillage'),
          create_ethanol_purification_system_after_beer_column.outs[1]]
)
def create_ethanol_purification_system(ins, outs,
                                       beer_column_heat_integration=True,
                                       IDs={}):
    beer, denaturant = ins
    ethanol, stillage, stripper_bottoms_product = outs
    distilled_beer = bst.Stream('')
    create_beer_distillation_system(
        ins=beer,
        outs=[distilled_beer, stillage],
        beer_column_heat_integration=beer_column_heat_integration,
        IDs=IDs,
        mockup=True,
    )
    create_ethanol_purification_system_after_beer_column(
        ins=[distilled_beer, denaturant],
        outs=[ethanol, stripper_bottoms_product],
        IDs=IDs,
        mockup=True,
    )
    

# %% Ethanol production section (fermentation and separations)

@SystemFactory(
    ID='sucrose_fermentation_sys',
    ins=[dict(ID='screened_juice', 
              Glucose=3802,
              Sucrose=4.309e+04,
              Water=2.59e+05,
              H3PO4=83.33,
              units='kg/hr',
              T=372)],
    outs=[dict(ID='beer'),
          dict(ID='evaporator_condensate'),
          dict(ID='vent')],
)
def create_sucrose_fermentation_system(ins, outs, scrubber=True):
    screened_juice, = ins
    beer, evaporator_condensate, vent = outs
    
    dilution_water = bst.Stream('dilution_water')
    
    ### Units ###
    
    # Split sugar solution
    S301 = units.Splitter('S301',
                          split=0.10)
    
    # Concentrate sugars
    F301 = units.MultiEffectEvaporator('F301',
                                       P=(101325, 69682, 47057, 30953, 19781),
                                       outs=('', evaporator_condensate),
                                       V_definition='First-effect',
                                       V=0.3) # fraction evaporated
    F301.specification = lambda: None
    # Note: value of steam ~ 6.86 for the following 
    # (101325, 73580.467, 50891.17, 32777.406, 19999.925, 11331.5),
    
    # Mix sugar solutions
    M301 = units.Mixer('M301')
    
    def get_sugar_concentration():
        s = M301.outs[0]
        if 'Lipid' in s.chemicals:
            ignored = float(s.imass['Lipid'])
        else:
            ignored = 0.
        return s.imass['Glucose', 'Sucrose'].sum() / (s.F_mass - ignored)
    
    def sugar_concentration_at_fraction_evaporated(V):
        F301.V = V
        for i in F301.path_until(M301): i._run()
        M301._run()
        return F301.sugar_concentration - get_sugar_concentration()
    
    def get_dilution_water():
        F301.V = 0
        for i in F301.path_until(M301): i._run()
        M301._run()
        target = F301.sugar_concentration
        current = get_sugar_concentration()
        return (1./target - 1./current) * sum([i.imass['Glucose', 'Sucrose'].sum() for i in M301.ins])
    
    @M301.add_specification(run=False)
    def adjust_glucose_concentration():
        V_guess = F301.V
        s_dilution_water = M301.ins[2]
        s_dilution_water.empty()
        dilution_water = get_dilution_water()
        if dilution_water < 0.:
            y1 = sugar_concentration_at_fraction_evaporated(1.)
            if y1 > 0.: 
                S301.split = 0.05
                S301._run()
                y1 = sugar_concentration_at_fraction_evaporated(1.)
                if y1 > 0.: 
                    S301.split = 0.
                    S301._run()
                    y1 = sugar_concentration_at_fraction_evaporated(1.)
                    if y1 > 0.:
                        raise RuntimeError('cannot evaporate to target sugar concentration')
            F301.V = flx.IQ_interpolation(
                sugar_concentration_at_fraction_evaporated,
                0., 1., x=V_guess, ytol=1e-5
            )
        else:
            s_dilution_water.imass['Water'] = dilution_water
        S301.split = 0.10
            
    F301.sugar_concentration = 0.23 # wt. % sugar
    
    # Cool for fermentation
    H301 = units.HXutility('H301', T=295.15)
    
    # Ethanol Production
    R301 = units.Fermentation('R301', outs=(('CO2' if scrubber else vent), ''), tau=9, efficiency=0.90, N=4) 
    T301 = units.StorageTank('T301', tau=4, vessel_material='Carbon steel')
    T301.line = 'Beer tank'
    
    if scrubber:
        stripping_water = bst.Stream('stripping_water',
                                     Water=26836,
                                     units='kg/hr')
        stripping_water_over_vent = stripping_water.mol / 21202.490455845436
        def update_stripping_water():
            stripping_water, vent = D301.ins
            stripping_water.mol[:] = stripping_water_over_vent * vent.F_mass
        
        D301 = units.VentScrubber('D301', ins=(stripping_water, R301-0), 
                                  outs=(vent, ''),
                                  gas=('CO2', 'O2'))
        M302 = units.Mixer('M302', outs=beer)
    
    # Separate 99% of yeast
    C301 = units.SolidsCentrifuge('C301', 
                                  outs=('', '' if scrubber else beer),
                                  split=(1-1e-6, 0.99, 1, 0.01),
                                  order=('Ethanol', 'Glucose', 'H3PO4', 'DryYeast'),
                                  solids=('DryYeast',))
    C301.split[:] = 1. - C301.split
    if 'Lipid' in C301.chemicals: C301.isplit['Lipid'] = 0.
    
    S302 = units.FakeSplitter('S302', None, ('', 'Yeast'))
    
    # Yeast mixing
    T305 = units.MixTank('T305')
    T305.tau = 0.1
    
    # Multi-effect evaporator pumps
    P306 = units.Pump('P306')
    
    def adjust_yeast_recycle():
        recycle, beer = S302.outs
        feed, = S302.ins
        yeast = 0.1 * feed.F_mass
        m = C301.moisture_content
        recycle.imass['Yeast', 'Water'] = [yeast, yeast * m / (1 - m)]
        beer.copy_like(feed)
        beer.separate_out(recycle, energy_balance=False)
        beer.mol[beer.mol < 0.] = 0.
        beer.T = recycle.T = feed.T
    S302.specification = adjust_yeast_recycle
    
    ### Ethanol system set-up ###
    
    screened_juice-S301-1-F301-0-P306
    (S301-0, P306-0, dilution_water)-M301-H301
    (H301-0, C301-0-S302-0)-R301-1-T301-0-C301-0-S302
    if scrubber: (C301-1, D301-1)-M302
    

@SystemFactory(
    ID='sucrose_to_ethanol_sys',
    ins=[create_sucrose_fermentation_system.ins[0],
         create_ethanol_purification_system.ins[1]], # denaturant
    outs=[*create_ethanol_purification_system.outs,
          create_sucrose_fermentation_system.outs[1]]
)
def create_sucrose_to_ethanol_system(ins, outs):
    screened_juice, denaturant = ins
    ethanol, stillage, stripper_bottoms_product, evaporator_condensate = outs
    
    beer = bst.Stream()
    
    create_sucrose_fermentation_system(
        ins=screened_juice,
        outs=[beer, evaporator_condensate],
        mockup=True
    )
    create_ethanol_purification_system(
        ins=[beer, denaturant], 
        outs=[ethanol, stillage, stripper_bottoms_product],
        mockup=True
    )
    
    

# %% Complete system

@SystemFactory(
    ID='sugarcane_sys', 
    ins=[*create_juicing_system_with_fiber_screener.ins,
         create_ethanol_purification_system.ins[1]], # denaturant
    outs=[create_ethanol_purification_system.outs[0], # ethanol
          dict(ID='vinasse'),
          dict(ID='wastewater'),
          dict(ID='emissions'),
          dict(ID='ash_disposal')]
)
def create_sugarcane_to_ethanol_system(ins, outs, use_area_convention=False):
    s = f.stream
    u = f.unit
    
    sugarcane, enzyme, H3PO4, lime, polymer, denaturant = ins
    ethanol, vinasse, wastewater, emissions, ash_disposal = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        area=100 if use_area_convention else None,
        ins=[sugarcane],
        outs=[''],
        mockup=True,
    )
    juicing_sys = create_juicing_system_with_fiber_screener(
        area=200 if use_area_convention else None,
        ins=[feedstock_handling_sys-0, enzyme, H3PO4, lime, polymer],
        mockup=True
    )
    ethanol_production_sys, edct = create_sucrose_to_ethanol_system(
        area=300 if use_area_convention else None,
        udct=True,
        ins=(juicing_sys-0, denaturant), outs=(ethanol, vinasse),
        mockup=True
    )
    M305 = units.Mixer(400 if use_area_convention else 'M305', 
        ins=(juicing_sys-2, *ethanol_production_sys-[2, 3]),
        outs=wastewater
    )
    
    ### Facilities ###    
    
    BT = units.BoilerTurbogenerator(400 if use_area_convention else 'BT',
        (juicing_sys-1, '', 'boiler_makeup_water', 'natural_gas', '', ''),
        outs=(emissions, 'rejected_water_and_blowdown', ash_disposal),
        boiler_efficiency=0.80,
        turbogenerator_efficiency=0.85
    )
    CT = units.CoolingTower(500 if use_area_convention else 'CT')
    makeup_water_streams = (s.cooling_tower_makeup_water,
                            s.boiler_makeup_water)
    process_water_streams = (s.imbibition_water,
                             s.rvf_wash_water,
                             s.stripping_water,
                             *makeup_water_streams)
    makeup_water = bst.Stream('makeup_water', price=0.000254)
    CWP = units.ChilledWaterPackage(500 if use_area_convention else 'CWP')
    PWC = units.ProcessWaterCenter(500 if use_area_convention else 'PWC',
                                   (bst.Stream(), makeup_water),
                                   (),
                                   None,
                                   makeup_water_streams,
                                   process_water_streams)
    
    F301 = edct['F301']
    D303 = edct['D303']
    HXN = bst.HeatExchangerNetwork(600 if use_area_convention else 'HXN',
                                   units=[F301, D303.condenser])
    
    # if vinasse_to_wastewater:
    #     plant_air = bst.Stream('plant_air', N2=83333, units='kg/hr')
    #     ADP = bst.facilities.AirDistributionPackage('ADP', plant_air)
    #     @ADP.add_specification(run=True)
    #     def adjust_plant_air():
    #         plant_air.imass['N2'] = 0.8 * feedstock_handling_sys.ins[0].F_mass
            
    #     wastewater_treatment_sys = bst.create_wastewater_treatment_system(
    #         ins=[vinasse],
    #         mockup=True,
    #     )
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

'''
References
----------
[1] Davis et al., Process Design and Economics for the Conversion of Lignocellulosic 
    Biomass to Hydrocarbon Fuels and Coproducts: 2018 Biochemical Design Case Update; 
    NREL/TP-5100-71949; National Renewable Energy Lab (NREL), 2018. 
    https://doi.org/10.2172/1483234
[2] Aden et al., Process Design Report for Stover Feedstock: Lignocellulosic
    Biomass to Ethanol Process Design and Economics Utilizing Co-Current Dilute
    Acid Prehydrolysis and Enzymatic Hydrolysis for Corn Stover; NREL/TP-510-32438;
    National Renewable Energy Lab (NREL), 2002.
    https://doi.org/10.2172/1218326.
[3] Argonne National Laboratory. The Greenhouse gases, Regulated Emissions,
    and Energy use in Transportation (GREET) Model https://greet.es.anl.gov/
    (accessed Aug 25, 2020).
[4] Roni et al., Herbaceous Feedstock 2018 State of Technology Report;
    INL/EXT-18-51654-Rev000; Idaho National Lab. (INL), 2020.
    https://doi.org/10.2172/1615147. 
[5] ecoinvent 3.6 https://www.ecoinvent.org/home.html (accessed Aug 26, 2020).

'''


# %%

# =============================================================================
# Setup
# =============================================================================

import biosteam as bst
import thermosteam as tmo
from ._chemicals import chems
from ._utils import auom

__all__ = ('price', 'CFs')


# %%

# =============================================================================
# Energy balances
# =============================================================================

_lps = bst.HeatUtility.get_heating_agent('low_pressure_steam')
_mps = bst.HeatUtility.get_heating_agent('medium_pressure_steam')
_hps = bst.HeatUtility.get_heating_agent('high_pressure_steam')
_mps.T = 233 + 273.15
_hps.T = 266 + 273.15

_cooling = bst.HeatUtility.get_cooling_agent('cooling_water')
_cooling.regeneration_price = 0
_cooling.T = 28 + 273.15
_cooling.T_limit = _cooling.T + 9

# Side steam in CHP not a heat utility, thus will cause problem in TEA utility
# cost calculation if price not set to 0 here, costs for regeneration of heating
# and cooling utilities will be considered as CAPEX and OPEX of CHP and CT, respectively
for i in (_lps, _mps, _hps, _cooling):
    i.heat_transfer_price = i.regeneration_price = 0


# %%

# =============================================================================
# Prices for techno-economic analysis (TEA), all in $/kg (electricity in $/kWh)
# and from ref [1] if not noted
# =============================================================================

bst.CE = 541.7 # year 2016
_kg_per_ton = auom('ton').conversion_factor('kg')
_lb_per_kg = auom('kg').conversion_factor('lb')
_liter_per_gallon = auom('gal').conversion_factor('L')
_ft3_per_m3 = auom('m3').conversion_factor('ft3')
_chemical_2020to2016 = 102.5 / 113.8 # average of Jan and Feb
_GDP_2007to2016 = 1.160

# From USD/dry-ton to USD/kg in 2016$, 20% moisture content,
# baseline and lower bound (60) from ref [1], upper bound (146.4) from
# Hartley et al., ACS Sustainable Chem. Eng. 2020, 8 (19), 7267–7277.
# https://doi.org/10.1021/acssuschemeng.9b06551.
_feedstock_factor = _kg_per_ton / 0.8
feedstock_price = 71.3 / _feedstock_factor
		
# Baseline from ref [1], lower and upper bounds (96% and 110% of baseline)
# calculated using the index of sulfuric acid from U.S. Bureau of Labor Statistics
# (Producer Price Index by Commodity for Chemicals and Allied Products)
# https://fred.stlouisfed.org/series/WPU0613020T1 (accessed Jul 31, 2020).
sulfuric_acid_price = 0.0430 * _lb_per_kg

# 2.2 is the average whole-sale ethanol price between 2010-2019 in 2016 $/gal 	
# based on Annual Energy Outlook (AEO) from Energy Information Adiministration (EIA)	
# (https://www.eia.gov/outlooks/aeo/), which is $0.732/gal and similar to the 	
# 2.2/(2988/1e3) = $0.736/gal based on a density of 2988 g/gal from H2 Tools
# Lower and upper bounds are $1.37/gal and $2.79/gal, or $0.460/kg and $0.978/kg
_ethanol_V = chems.Ethanol.V('l', 298.15, 101325) # molar volume in m3/mol	
_ethanol_MW = chems.Ethanol.MW
_ethanol_kg_2_gal = _liter_per_gallon/_ethanol_V*_ethanol_MW/1e6
ethanol_price = 2.2 / _ethanol_kg_2_gal

# Cost is negative because it's a product stream
ash_disposal_price = -1.41e6 / (4279*7880)

# Assums no cost/credit for baseline, the same as ash disposal for the lower bound,	
# for the upper bound (i.e., positive selling price indicating profit), use 	
# USGS 2015-2019 average  free on bard price in $/metric tonne for crude gypsum. 
# National Minerals Information Center. Mineral Commodity Summaries 2020; 
# U.S. Geological Survey, 2020.
# Assuming all prices were in their nominal year (e.g., 2015 price in 2015$)
# and adjusted to 2016$
# 2015: 7.80 * 1.114 / 1.100 = 7.90
# 2016: 8.00
# 2017: 7.50 * 1.114 / 1.134 = 7.37
# 2018: 8.30 * 1.114 / 1.157 = 7.99
# 2019: 8.00 * 1.114 / 1.185 = 7.52
# (7.90+8.00+7.37+7.99+7.52) / 5 = 7.76 (in metric tonne)
# For the lower bound (i.e., negative selling price indicating cost), use price from
# ref [2]: $0.0094/lb in 2000$ = 0.0094*1.114/0.802*2.20462 = $0.0288/kg in 2016$
gypsum_price = 0

# Baseline from ref [1], lower bound is 2015-2019 average of 	
# hydrate lime in $/ton at plant from Mineral Commodity Summaries 2020.	
# 2015: 146.40 * (1.114/1.100) / 907.18474 = 0.163	
# 2016: 145.50 / 907.18474 = 0.160	
# 2017: 147.10 * (1.114/1.134) / 907.18474 = 0.159	
# 2018: 151.50 * (1.114/1.157) / 907.18474 = 0.161	
# 2019: 151.00 * (1.114/1.185) / 907.18474 = 0.156	
# (0.163+0.160+0.159+0.161+0.156) / 5 = 0.160	
# Upper bound is +10% from baseline = 0.1189 * _lb_per_kg * 1.1 = 0.288
lime_price = 0.1189 * _lb_per_kg

baghouse_bag_price = 466183/5/(24*365*0.96) * _GDP_2007to2016

# 4.70 is the average natural gas price in 2016$/Mcf based on AEO from EIA,
# which is $0.231/kg at 273.15 K or $0.253/kg at 298.15 K using BioSTEAM,   
# similar to the 4.7/1000/22*1000 = $0.214/kg at 273.15 K using 22 g/ft3 from H2 Tools
# Using the same conversion, lower and upper bounds (min/max of 2010-2019) should be 
# $3.68/Mcf and $5.65/Mcf, or $0.198/kg and $0.304/kg
_CH4_V = chems.CH4.V(298.15, 101325) # molar volume in m3/mol
_CH4_MW = chems.CH4.MW
natural_gas_price = 4.70/1e3*_ft3_per_m3*_CH4_V * (1e3/_CH4_MW)

# https://www.rightpricechemicals.com/buy-amberlyst-15-ion-exchange-resin.html	
# USD 383.13 for 2.5kg (largest available size order), accessed 06/11/2020
amberlyst_15_price = 153.252 * _chemical_2020to2016

# All in 2016$/kg
price = {'Feedstock': feedstock_price, 	
         'H2SO4': 0.0430 * _lb_per_kg,
         # 0.1900 is for NH3	
         'NH4OH': 0.1900 * _lb_per_kg * chems.NH3.MW/chems.NH4OH.MW,
         'CSL': 0.0339 * _lb_per_kg,
         'Enzyme': 6.16,
         'Lime': lime_price,
         'Amberlyst15': amberlyst_15_price,
         'NaOH': 0.2384 * _lb_per_kg,
         'WWT polymer': 2.6282 * _lb_per_kg,
         'Boiler chems': 2.9772 * _lb_per_kg,	
         'Cooling tower chems': 1.7842 * _lb_per_kg,	
         'Makeup water': 0.0002 * _lb_per_kg,	
         # Cost of ash is negative because it's a product stream	
         'Ash disposal': ash_disposal_price,	
         'Gypsum': gypsum_price,
         'Ethanol': ethanol_price,	
         'Baghouse bag': baghouse_bag_price,	
         'Natural gas': natural_gas_price}

# $/kWh, from EIA AEO, 2010-2019 average (0.067-0.074 range)
bst.PowerUtility.price = 0.070


# %%

# =============================================================================
# Characterization factors (CFs) for life cycle analysis (LCA), all from ref [3] 
# if not noted, note that it is unclear if in-plant receiving and preprocessing
# (~50% of the total impact per ref [4]) of feedstock is included in ref [3]
# =============================================================================

CFs = {}

# =============================================================================
# 100-year global warming potential (GWP) in kg CO2-eq/kg
# =============================================================================
GWP_CFs = {
    'NH4OH': 2.64 * chems.NH3.MW/chems.NH4OH.MW,
    'CSL': 1.55,
    'CH4': 0.40, # NA NG from shale and conventional recovery
    'Enzyme': 2.24,
    'Lime': 1.29 * 56.0774/74.093, # CaO to Ca(OH)2
    'NaOH': 2.11,
    'H2SO4': 44.47/1e3,
    'Ethanol': 1.44
    }

GWP_CF_array = chems.kwarray(GWP_CFs)
# In kg CO2-eq/kg of material
GWP_CF_stream = tmo.Stream('GWP_CF_stream', GWP_CF_array, units='kg/hr')

GWP_CFs['CaCO3'] = 10.30/1e3
GWP_CFs['Gypsum'] = -4.20/1e3
# In kg CO2-eq/kWh
GWP_CFs['Electricity'] = 0.48
# Lactic acid from corn stover
GWP_CFs['Lactic acid_GREET'] = 1.80
# From ref [5], lactic acid production, RoW, TRACI global warming,
# substitution, consequential, long-term
GWP_CFs['Lactic acid_fossil'] = 5.2796

CFs['GWP_CFs'] = GWP_CFs
CFs['GWP_CF_stream'] = GWP_CF_stream

# =============================================================================
# Fossil energy consumption (FEC), in MJ/kg of material
# =============================================================================

FEC_CFs = {
    'NH4OH': 42 * chems.NH3.MW/chems.NH4OH.MW,
    'CSL': 12,
    'CH4': 50, # NA NG from shale and conventional recovery
    'Enzyme': 26,
    'Lime': 4.896 * 56.0774/74.093, # CaO to Ca(OH)2
    'NaOH': 29,
    'H2SO4': 568.98/1e3,
    'Ethanol': 16
    }

FEC_CF_array = chems.kwarray(FEC_CFs)
# In MJ/kg of material
FEC_CF_stream = tmo.Stream('FEC_CF_stream', FEC_CF_array, units='kg/hr')

FEC_CFs['CaCO3'] = 133.19/1e3
FEC_CFs['Gypsum'] = -44.19/1e3
# In MJ/kWh
FEC_CFs['Electricity'] = 5.926
# From corn stover
FEC_CFs['Lactic acid'] = 29
# From ref [5], lactic acid production, RoW, cumulative energy demand, fossil,
# substitution, consequential, long-term
FEC_CFs['Lactic acid_fossil'] = 91.265

CFs['FEC_CFs'] = FEC_CFs
CFs['FEC_CF_stream'] = FEC_CF_stream













# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from biorefineries.lipidcane.system import lipidcane_sys, lipidcane_tea, lipidcane, makeup_water
from biorefineries.lipidcane.utils import set_lipid_fraction
from biosteam.plots import MetricBar, CABBI_green_colormap, plot_contour_2d
import matplotlib.pyplot as plt
import numpy as np


def evaluate_feedstock_metrics(lipid_content, plant_size, operating_days):
    set_lipid_fraction(lipid_content)
    lipidcane_tea.operating_days = operating_days
    lipidcane.F_mass = plant_size / 24. / operating_days * 907.18474
    lipidcane_sys.simulate()
    feedstock_price = lipidcane_tea.solve_price(lipidcane) * 907.18474 # USD / ton
    water_consumption = makeup_water.get_total_flow('gal/day') * operating_days / 1e6 # MMGal / yr
    installed_cost = lipidcane_tea.installed_cost / 1e6 # million USD
    return np.array([feedstock_price, water_consumption, installed_cost])

evaluate_feedstock_metrics = np.vectorize(evaluate_feedstock_metrics,
                                          signature="(),(),()->(3)")

# %% Evaluate data

N_points = 10
lipidcane_tea.IRR = 0.10
lipid_content_lb = 0.01
lipid_content_ub = 0.15
lipid_content = np.linspace(lipid_content_lb, lipid_content_ub, N_points)
plant_size_lb = 5e5
plant_size_ub = 4 * plant_size_lb
plant_size = np.linspace(plant_size_lb, plant_size_ub, N_points)
operating_days_lb = 100
operating_days_ub = 350
operating_days_1d = np.linspace(operating_days_lb, operating_days_ub, 3)
lipid_content, plant_size, operating_days = np.meshgrid(lipid_content, plant_size, operating_days_1d)
data = evaluate_feedstock_metrics(lipid_content, plant_size, operating_days)


# %% Plot

xlabel = r'Lipid content [wt. %]'
ylabel = r"Plant size [$\mathrm{MMTon} \cdot \mathrm{yr}^{-1}$]"
xticks = [1, 3, 5, 7, 9, 11, 13, 15]
yticks = np.array([0.5  , 0.875, 1.25 , 1.625, 2.   ])
million_dollar = r"\mathrm{\$} \cdot \mathrm{10}^{6}"
MFP_units = r"$\mathrm{\$} \cdot \mathrm{ton}^{-1}$"
Water_units = r"$\mathrm{MMGal} \cdot \mathrm{yr}^{-1}$"
operating_days_units = r"$\mathrm{days} \cdot \mathrm{yr}^{-1}$"
installed_cost_units = f"${million_dollar}$"
# VOC_units = "$" + million_dollar + r"\cdot \mathrm{yr}^{-1}$"
# installed_cost_units = f"${million_dollar}$"
metric_bars = (MetricBar('MFP', MFP_units, CABBI_green_colormap(), 
                         [0, 15, 30, 45, 60]),
               MetricBar('Water use\n', Water_units, plt.cm.get_cmap('bone_r'),
                         [0, 50, 100, 150, 200]),
               MetricBar('Inst. cost\n', installed_cost_units, plt.cm.get_cmap('magma_r'), 
                         [0, 75, 150, 225, 300]))
plot_contour_2d(100. * lipid_content[:, :, 0],
                plant_size[:, :, 0] / 1e6, operating_days_1d, np.swapaxes(data, 2, 3), 
                xlabel, ylabel, xticks, yticks, metric_bars, fillblack=True,
                Z_label="Operation",
                Z_value_format=lambda Z: f"{Z:.0f} {operating_days_units}")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

from biorefineries.cornstover import CellulosicEthanolTEA

__all__ = ('LacticTEA',)

class LacticTEA(CellulosicEthanolTEA):
    
    # For uncertainty analysis
    _TCI_ratio_cached = 1








# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module defines the composition_balance function, which performs mass 
energy balance the given oil composition of lipid cane to determine its
composition.

"""
from thermosteam import Stream
from biosteam import main_flowsheet as f
from numpy import array

__all__ = ('set_lipid_fraction',
           'get_lipid_fraction')

def set_lipid_fraction(lipid_fraction, stream=None,
                       PL_fraction=0.,
                       FFA_fraction=0.,
                       z_mass_carbs_baseline=0.149,
                       z_mass_ash_baseline=0.006,
                       z_mass_solids_baseline=0.015):
    """Adjust composition of lipid cane to achieve desired oil fraction (dry weight)."""
    if not stream: stream = f.stream.lipidcane
    thermo = stream.thermo
    carbs_IDs = ('Glucose', 'Sucrose')
    fiber_IDs = ('Lignin', 'Cellulose', 'Hemicellulose')
    lipid_IDs = tuple([i for i in ('PL', 'FFA', 'MAG', 'DAG', 'TAG') if i in thermo.chemicals])
    carbs = Stream(None, thermo=thermo)
    fiber = Stream(None, thermo=thermo)
    lipid = Stream(None, TAG=1.0 - PL_fraction - FFA_fraction, units='kg/hr', thermo=thermo)
    if PL_fraction: lipid.imass['PL'] = PL_fraction
    if FFA_fraction: lipid.imass['FFA'] = FFA_fraction
    
    carbs.imol[carbs_IDs] = stream.imol[carbs_IDs]
    fiber.imol[fiber_IDs] = stream.imol[fiber_IDs]
    
    # Mass property arrays
    carbs_mass = stream.imass[carbs_IDs].value
    fiber_mass = stream.imass[fiber_IDs].value
    lipid_mass = lipid.imass[lipid_IDs].value
    
    
    # Net weight
    carbs_massnet = carbs_mass.sum()
    fiber_massnet = fiber_mass.sum()
    lipid_massnet = lipid_mass.sum()
    
    # Heats of combustion per kg
    LHV_carbs_kg = carbs.LHV / carbs_massnet
    LHV_lipid_kg = lipid.LHV / lipid_massnet
    LHV_lipid_over_carbs = LHV_lipid_kg / LHV_carbs_kg
    
    # Relative composition
    r_mass_lipid = lipid_mass / lipid_massnet
    r_mass_carbs = carbs_mass / carbs_massnet
    r_mass_fiber = fiber_mass / fiber_massnet
    F_mass = stream.F_mass
    z_mass_lipid = lipid_fraction
    z_dry = 0.3
    z_mass_lipid = z_mass_lipid * z_dry 
    z_mass_carbs = z_mass_carbs_baseline - z_mass_lipid * LHV_lipid_over_carbs
    z_mass_fiber = z_dry - z_mass_carbs - z_mass_lipid - z_mass_ash_baseline - z_mass_solids_baseline
    imass = stream.imass
    imass['Water'] = F_mass * 0.7
    imass['Ash'] = F_mass * z_mass_ash_baseline
    imass['Solids'] = F_mass * z_mass_solids_baseline
    imass[lipid_IDs] = r_mass_lipid * z_mass_lipid * F_mass
    imass[carbs_IDs] = r_mass_carbs * z_mass_carbs * F_mass
    imass[fiber_IDs] = r_mass_fiber * z_mass_fiber * F_mass
    if any(stream.mol < 0):
        raise ValueError(f'lipid cane oil composition of {z_mass_lipid/z_dry*100:.0f}% dry weight is infeasible')

def get_lipid_fraction(stream=None):
    if not stream: stream = f.stream.lipidcane
    imass = stream.imass
    F_dry_mass = stream.F_mass - imass['Water']
    return imass['Lipid'] / F_dry_mass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %% 

# =============================================================================
# Setup
# =============================================================================

import numpy as np
import pandas as pd
import biosteam as bst
from warnings import warn
from biosteam.utils import TicToc
from biorefineries.lactic import load_system, \
    SSCF_flowsheet, SSCF_funcs, SHF_flowsheet, SHF_funcs
from biorefineries.lactic._chemicals import sugars
from biorefineries.lactic._utils import set_yield

load_system('SSCF')
load_system('SHF')


# %% 

# =============================================================================
# Evaluate system at different lactic acid titer and yield (conversion),
# using either regular strain (need lime addition during fermentation to neutralize
# the produced lactic acid) or acid-resistant strain (no neutralization need)
# =============================================================================

# Initiate a timer
timer = TicToc('timer')
timer.tic()
run_number = 0
sugar_concs = [[], []]
lactics = {'target yield': [],
           'target titer': [],
           'actual yield': [],
           'actual titer': []}
MPSPs = {0.89: [],
         0.18: [],
         1.92: []}
NPVs = {0.89: [],
        0.18: [],
        1.92: []}
GWPs = {0.89: [],
        0.18: [],
        1.92: []}
FECs = {0.89: [],
        0.18: [],
        1.92: []}

def compute_sugar_conc(stream):
    return stream.imass[sugars].sum()/stream.F_vol

def solve_TEA(lactic_acid, lactic_tea):
    lactic_acid.price = 0
    for i in range(3):
        MPSP = lactic_acid.price = lactic_tea.solve_price(lactic_acid)
    return MPSP

def update_productivity(R301, R302, productivity):
    R301.productivity = productivity
    R302.productivity = productivity * R302.ferm_ratio
    for unit in (R301, R302):
        unit._design()
        unit._cost()

def simulate_log_results(kind):
    kind = kind.upper()
    if 'SSCF' in kind:
        flowsheet = SSCF_flowsheet
        funcs = SSCF_funcs
    elif 'SHF' in kind:
        flowsheet = SHF_flowsheet
        funcs = SHF_funcs
    bst.main_flowsheet.set_flowsheet(flowsheet)
    load_system(kind)
    
    R301 = flowsheet.unit.R301
    R302 = flowsheet.unit.R302
    lactic_acid = flowsheet.stream.lactic_acid
    lactic_sys = flowsheet.system.lactic_sys
    lactic_tea = lactic_sys.TEA
    
    try:
        lactic_sys.simulate()
        for productivity in (0.89, 0.18, 1.92):
            update_productivity(R301, R302, productivity)
            MPSP = solve_TEA(lactic_acid, lactic_tea)
            GWP = funcs['get_GWP']()
            FEC = funcs['get_FEC']()
            MPSPs[productivity].append(MPSP)
            GWPs[productivity].append(GWP)
            FECs[productivity].append(FEC)
            NPVs[productivity].append(lactic_tea.NPV)
        lactics['actual yield'].append(R301.lactic_yield)
        lactics['actual titer'].append(R301.effluent_titer)
    except:
        warn(f'Simulation failed at target yield {round(R301.target_yield,2)}, ' \
              f'target titer {R301.target_titer}.')
        lactic_sys.empty_recycles()
        lactic_sys.reset_cache()
        lactics['actual yield'].append(np.nan)
        lactics['actual titer'].append(np.nan)
        for m in (MPSPs, NPVs, GWPs, FECs):
            for n in m.keys():
                m[n].append(np.nan)

    global run_number
    run_number += 1
    print(f'Target yield is {round(R301.target_yield,2)}, actual yield is {round(R301.cofermentation_rxns.X[0], 2)}; ' \
          f'target titer is {R301.target_titer}, actual titer is {round(R301.effluent_titer)}.')


def save_data_clear():
    df = pd.DataFrame({
        ('Lactic acid', 'Target Yield [g/g]'): lactics['target yield'],
        ('Lactic acid', 'Target Titer [g/L]'): lactics['target titer'],
        ('Lactic acid', 'Actual Yield [g/g]'): lactics['actual yield'],
        ('Lactic acid', 'Actual Titer [g/L]'): lactics['actual titer'],
        ('0.89 [g/L/hr]', 'MPSP [$/kg]'): MPSPs[0.89],
        ('0.89 [g/L/hr]', 'NPV [$]'): NPVs[0.89],
        ('0.89 [g/L/hr]', 'GWP [kg CO2-eq/kg]'): GWPs[0.89],
        ('0.89 [g/L/hr]', 'FEC [MJ/kg]'): FECs[0.89],
        ('0.18 [g/L/hr]', 'MPSP [$/kg]'): MPSPs[0.18],
        ('0.18 [g/L/hr]', 'NPV [$]'): NPVs[0.18],
        ('0.18 [g/L/hr]', 'GWP [kg CO2-eq/kg]'): GWPs[0.18],
        ('0.18 [g/L/hr]', 'FEC [MJ/kg]'): FECs[0.18],
        ('1.92 [g/L/hr]', 'MPSP [$/kg]'): MPSPs[1.92],
        ('1.92 [g/L/hr]', 'NPV [$]'): NPVs[1.92],
        ('1.92 [g/L/hr]', 'GWP [kg CO2-eq/kg]'): GWPs[1.92],
        ('1.92 [g/L/hr]', 'FEC [MJ/kg]'): FECs[1.92]
        })
    for i in (lactics, MPSPs, NPVs, GWPs, FECs):
        for j in i.keys():
            i[j] = []
    return df

def run_TRY(yield_range, kind, mode, feed_freq, if_resistant, titer_range):
    bst.speed_up()
    if 'SSCF' in str(kind).upper():
        flowsheet = SSCF_flowsheet
    elif 'SHF' in str(kind).upper():
        flowsheet = SHF_flowsheet
    bst.main_flowsheet.set_flowsheet(flowsheet)

    u = flowsheet.unit
    R301 = u.R301
    R302 = u.R302
    R401 = u.R401
    S402 = u.S402
    R301.mode = mode
    R301.feed_freq = feed_freq
    R301.allow_dilution = True
    if hasattr(R301, 'allow_concentration'):
        R301.allow_concentration = True
    R401.bypass = if_resistant
    S402.bypass = if_resistant
    for i in yield_range:
        for j in titer_range:
            lactics['target yield'].append(i)
            lactics['target titer'].append(j)
            R301.target_yield = i
            R301.target_titer = j
            set_yield(i, R301, R302)
            simulate_log_results(kind)
            if R301.effluent_titer+2<j:
                n_remained = len(titer_range) - titer_range.index(j) - 1
                lactics['target yield'].extend([i]*n_remained)
                lactics['target titer'].extend(titer_range[titer_range.index(j)+1:])
                lactics['actual yield'].extend([i]*n_remained)
                lactics['actual titer'].extend(['']*n_remained)
                for p in (0.89, 0.18, 1.92):
                    for m in (MPSPs, NPVs,GWPs, FECs):
                        m[p].extend(['']*n_remained)
                break
    
    print(f'\nSimulation time is {timer.elapsed_time/60:.1f} min for {run_number} runs.')


# %%

# =============================================================================
# Regular strain
# =============================================================================

# Change this to increase resolution
yield_range = np.arange(0.3, 1.01, 0.2) - 1e-6
# yield_range = np.arange(0.3, 1.01, 0.025) - 1e-6
yield_range = yield_range.tolist()

# Change this to increase resolution
titer_range = np.arange(50, 220, 50)
# titer_range = np.arange(50, 250, 1)
titer_range = titer_range.tolist()

print('\n---------- SSCF Regular Strain Batch Mode ----------')
run_TRY(yield_range=yield_range, kind='SSCF', mode='batch', feed_freq=1,
        if_resistant=False, titer_range=titer_range)
SSCF_reg_b = save_data_clear()
SSCF_reg_b.to_excel('SSCF_reg_batch.xlsx')

# Change this for fed-batch mode
# for i in (1, 3, 5, 10):
for i in (1,):
    print(f'\n---------- SHF Regular Strain Batch Feed {i+1} Times ----------')
    run_TRY(yield_range=yield_range, kind='SHF', mode='batch', feed_freq=i+1,
            if_resistant=False, titer_range=titer_range)
    SHF_reg_b = save_data_clear()
    SHF_reg_b.to_excel(f'SHF_reg_batch{i+1}.xlsx')

print('\n---------- SHF Regular Strain Continuous ----------')
run_TRY(yield_range=yield_range, kind='SHF', mode='continuous', feed_freq=1,
        if_resistant=False, titer_range=titer_range)
SHF_reg_c = save_data_clear()
SHF_reg_c.to_excel('SHF_reg_continuous.xlsx')


# %%

# =============================================================================
# Acid-resistant strain
# =============================================================================

print('\n---------- SSCF Regular Strain Batch Mode ----------')
run_TRY(yield_range=yield_range, kind='SSCF', mode='batch', feed_freq=1,
        if_resistant=True, titer_range=titer_range)
SSCF_reg_b = save_data_clear()
SSCF_reg_b.to_excel('SSCF_reg_batch.xlsx')

# Change this for fed-batch mode
# for i in (1, 3, 5, 10):
for i in (1,):
    print(f'\n---------- SHF Regular Strain Batch Feed {i} Times ----------')
    run_TRY(yield_range=yield_range, kind='SHF', mode='batch', feed_freq=i,
            if_resistant=True, titer_range=titer_range)
    SHF_reg_b = save_data_clear()
    SHF_reg_b.to_excel(f'SHF_reg_batch{i+1}.xlsx')

print('\n---------- SHF Regular Strain Continuous ----------')
run_TRY(yield_range=yield_range, kind='SHF', mode='continuous', feed_freq=1,
        if_resistant=True, titer_range=titer_range)
SHF_reg_c = save_data_clear()
SHF_reg_c.to_excel('SHF_reg_continuous.xlsx')





# -*- coding: utf-8 -*-
"""
Created on Tue Apr 28 01:25:56 2020

@author: yoelr
"""

from . import fatty_alcohol_bioreactor
from . import slle_centrifuge
from biorefineries.cornstover.units import DAPStorageTank, CSLStorageTank 

__all__ = (*fatty_alcohol_bioreactor.__all__,
           *slle_centrifuge.__all__,
           'DAPStorageTank', 'CSLStorageTank',
)

from .fatty_alcohol_bioreactor import *
from .slle_centrifuge import *



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 13 21:43:56 2019

@author: yoelr
"""
from biosteam.evaluation import evaluation_tools as tools
from biosteam.evaluation import Model, Metric
from biorefineries.cornstover import \
    cornstover_sys, cornstover_tea, \
    ethanol, cornstover, R301, R303, ethanol_density_kggal, \
    areas, BT, Area700, AllAreas

cornstover_sys.simulate()
get_MESP = lambda: cornstover_tea.solve_price(ethanol) * ethanol_density_kggal
get_FCI = lambda: cornstover_tea.FCI
get_coproduct_credit = lambda: cornstover_tea.utility_cost
get_ethanol_production = lambda: ethanol.F_mass
get_steam_demand = lambda: BT.steam_demand.F_mass
get_electricity_demand = AllAreas.get_electricity_consumption
get_electricity_production = AllAreas.get_electricity_production
get_excess_electricity = lambda: get_electricity_production() - get_electricity_demand()

metrics =[Metric('Minimum ethanol selling price', get_MESP, 'USD/gal'),
          Metric('Fixed capital investment', get_FCI, 'USD'),
          Metric('Co-product credit', get_coproduct_credit, 'USD/yr'),
          Metric('Ethanol production', get_ethanol_production, 'kg/hr'),
          Metric('Steam demand', get_steam_demand, 'kg/hr'),
          Metric('Excess electricity', get_excess_electricity, 'MW'),
          Metric('Electricity production', get_electricity_production, 'MW'),
          Metric('Total electricity demand', get_electricity_demand, 'MW')]

for i, area in enumerate(areas, 1):
    Area = f'Area {i}00'
    metrics.extend(
        (Metric('Electricity', area.get_electricity_consumption, 'MW', Area),
         Metric('Cooling duty', area.get_cooling_duty, 'GJ/hr', Area),
         Metric('Installed equipment cost', area.get_installed_cost, '10^6 USD', Area)))

cornstover_model = Model(cornstover_sys, metrics)
cornstover_model.load_default_parameters(cornstover, operating_days=False)
cornstover_sys.simulate()
param = cornstover_model.parameter

# Add saccharification as a parameter
saccharification_reaction = R301.saccharification[2]
X = tools.bounded_triang(saccharification_reaction.X, addition=0.04)
@param(element=R301, kind='coupled', distribution=X, baseline=saccharification_reaction.X)
def set_saccharification_conversion(saccharification_conversion):
    saccharification_reaction.X = saccharification_conversion

# Add ethanol conversion as a parameter
ethanol_reaction = R303.cofermentation[0]
X = tools.bounded_triang(ethanol_reaction.X, addition=0.02)
@param(element=R301, kind='coupled', distribution=X, baseline=ethanol_reaction.X)
def set_ethanol_conversion(ethanol_conversion):
    ethanol_reaction.X = ethanol_conversion
    
# Add saccharification time as a parameter
X = tools.triang(R303.tau_saccharification)
@param(element=R303, kind='isolated', distribution=X, baseline=R303.tau_saccharification)
def set_saccharification_time(saccharification_time):
    R303.tau_saccharification= saccharification_time
    
# Add fermentation time as a parameter
X = tools.triang(R303.tau_cofermentation)
@param(element=R303, kind='isolated',  distribution=X, baseline=R303.tau_cofermentation)
def set_fermentation_time(fermentation_time):
    R303.tau_cofermentation = fermentation_time
